# Functions for performing randomizaiton tests on ts
# Author: Iavor Bojinov <iav.bojinov@gmail.com>


est.ht <- function(Yobs, Wobs, pT, p.lag = 0, q.step = 0, normalize = FALSE) {
  # Description:
  #   This function computes the p-lag q-step estimator of the p-lag q-step 
  #   causal effect. p.lag = 0 corresponds to the contemporaneous causal effect.
  # Args:
  #   Yobs: The observed data.
  #   Wobs: The corresponding treatment assignments.
  #   pT: Either a number giving the probability of receiving treatment at each
  #       time point, or a vector containing the probability of receiving
  #       treatment at each time poit.
  #   p.lag: A non-negative integer indicating the p-lag.
  #   q.step: A positive integer; indicating the number of steps.
  #   normalize: A logical value, indicating if we should compute the normalize
  #              version of the step estimator using the fisher variance. 
  # Returns:
  #   est.mean: The mean of the estimated values. 
  #   est: The estimated value

  stopifnot((q.step + p.lag + 1) > 0)
  p.lag <- p.lag + 1
  T <- length(Wobs)
  if (class(Wobs) == "factor"){
    # relable the factor levels
    levels(Wobs) <- c(0, 1)
    Wobs <- as.numeric(Wobs) - 1
  }
  if (length(pT) == 1) {
    pT.temp <- rep(pT, T)
    pT.temp[Wobs == 0] <- 1 - pT
    pT <- pT.temp
  }
  # Compute the probability of the received treatment
  if ((q.step + p.lag) > 1) {
    pT.mat <- matrix(NA, nrow = (q.step + p.lag), ncol = T)
    pT.mat[1, ] <- pT
    for (kk in 1:(q.step + p.lag - 1)) {
      pT.mat[(kk + 1), ] <- c(rep(1, kk), pT[1 : (T - kk)]) 
    }
    pT.prod <- apply(pT.mat, 2, prod)
  } else {
    pT.prod <- pT
  }
  if (p.lag > 0){
    Wobs <- c(rep(0, p.lag - 1), Wobs[1:(T - p.lag + 1)])
  }
  est <- as.numeric(Yobs * (( - 1)^(1 - Wobs)) * ((1 / 2) ^ 
                   ((p.lag + q.step) - 1)) / pT.prod)
  if (normalize) {
    gm <- var.fisher(Yobs, Wobs, pT, p.lag, q.step)$var.est
    # Sometimes with very low probability gm = 0, then you just add some noise 
    # since 0/a = 0, this doesn't change anything 
    gm[which(gm == 0)] <- abs(jitter(gm[which(gm == 0)]))
    est <- est / sqrt(gm)
  }
  est.mean <- mean(est[(p.lag + q.step):T], na.rm = TRUE)
  return(list(est.mean = est.mean, est = est))
}


p.value.rand <- function(Yobs, Wobs, pT, p.lag = 0, q.step = 0, N.it = 1000, 
                         null.val = 0, normalize = FALSE) {
  # Description:
  #   This functions takes an observed outcome path and treatment path, and
  #   computes a randomization based p-value for the ATE = delta.
  # Args:
  #   Yobs: The observed data.
  #   Wobs: The corresponding treatment assignments.
  #   pT: Either a number giving the probability of receiving treatment at each
  #       time point, or a vector containing the probability of receiving
  #       treatment at each time poit.
  #   p.lag: A non-negative integer indicating the p-lag.
  #   q.step: A positive integer; indicating the number of steps.
  #   normalize: A logical value, indicating if we should compute the normalize
  #              version of the step estimator using the fisher variance. 
  #   null.val: The null value we wish to test against, this is used to build CI
  #   N: A numeric, the number of randomizations used to generate the p-value.
  # Returns:
  #   p.val: The estimated p-value.
  #   rand.dist: The randomization distribution. 

  # compute the initial estimate
  est <- est.ht(Yobs, Wobs, pT, p.lag, q.step, normalize)$est.mean
  rand.dist <- rep(NA, N.it)
  p.val <- 0
  T <- length(Yobs)
  for(kk in 1:N.it) {
    cat("Computing iteration: ")
    cat(kk)
    Wnew <- rbinom(n = T, size = 1, prob = pT)
    est.t <- est.ht(Yobs, Wnew, pT, p.lag, q.step, normalize)
    rand.dist[kk] <- est.t$est.mean + null.val
    p.val <- p.val + ifelse(abs(rand.dist[kk]) >= abs(est), 1, 0)
    cat("\r")
  }
  p.val <- p.val / N.it
  return(list(est = est, p.val = p.val, rand.dist = rand.dist))
}

p.value.norm <- function(Yobs, Wobs, pT, p.lag = 0, q.step = 0, rCI = FALSE,
                         ap = 0.05) {
  # Description:
  #   This functions takes an observed outcome path and treatment path, and
  #   computes conservative p-value as well as a (1-ap)% confidence interval 
  #   (whenever rCI == TRUE). 
  # Args:
  #   Yobs: A vector of length T, the observed outcome path.
  #   Wobs: A vector of treatment assignments
  #   pT: A numeric, the probability of assigning a unit to treatment.
  #   p.lag: A non-negative integer indicating the p-lag.
  #   q.step: A positive integer; indicating the number of steps.
  #   rCI: A logical value indicating if a 95% normal CI should be returned
  #   ap: Significance level for the confidence interval.
  # Returns:
  #   Z: The Z statistic.
  #   p.val: The estimated p-value.

  T <- length(Yobs)
  if (length(pT) == 1) {
    pT.temp <- rep(pT, T)
    pT.temp[Wobs == 0] <- 1 - pT
    pT <- pT.temp
  }
  est <- est.ht(Yobs, Wobs, pT, p.lag, q.step, normalize = FALSE)$est.mean
  sigma <- sqrt(var.bound(Yobs, Wobs, pT, p.lag, q.step)$gm.sq.est)
  Z <- est / sigma
  p.val <- 2 * min(1 - pnorm(Z), pnorm(Z))
  if (rCI) {
    CI <- est + c(sigma * qnorm(1 - ap / 2), - sigma * qnorm(1 - ap / 2))
    out <- list(est = est, Z = Z, p.val = p.val, CI = CI)
  } else {
    out <- list(est = est, Z = Z, p.val = p.val)
  }
  return(out)
}

var.bound <- function(Yobs, Wobs, pT, p.lag, q.step) {
  # Description:
  #   This functions takes an observed outcome path and treatment path, and
  #   computes the variance bound given in (14)
  # Args:
  #   Yobs: A vector of length T, the observed outcome path.
  #   Wobs: A vector of treatment assignments
  #   pT: A numeric, the probability of assigning a unit to treatment.
  #   p.lag: A non-negative integer indicating the p-lag.
  #   q.step: A positive integer; indicating the number of steps.
  # Returns:
  #   var.est: The estimated variance bound at each time point.
  #   gm.sq.est: The estimate of gm (the average of the variances). 
  T <- length(Wobs)
  if (class(Wobs) == "factor"){
    # relable the factor levels
    levels(Wobs) <- c(0, 1)
    Wobs <- as.numeric(Wobs) - 1
  }
  K.step <- p.lag + q.step + 1
  # Turn pT into a vector
  if (length(pT) == 1) {
    pT.temp <- rep(pT, T)
    pT.temp[Wobs == 0] <- 1 - pT
    pT <- pT.temp
  }
  if (K.step > 1) {
    pT.mat <- matrix(NA, nrow = K.step, ncol = T)
    pT.mat[1, ] <- pT * Wobs + (1 - pT) * (1 - Wobs)
    for (kk in 1:(K.step - 1)) {
      pT.mat[(kk + 1), ] <- c(rep(1, kk), pT.mat[1, 1:(T - kk)]) 
    }
    pT.prod <- apply(pT.mat, 2, prod)
  } else {
    pT.prod <- pT * Wobs + (1 - pT) * (1 - Wobs)
  }
  var.est <- as.numeric(Yobs^2  * (1 + 2 * pT.prod * (2^(K.step - 1) - 1)) *
                        ((1 / 4)^(K.step - 1)) / pT.prod^2) 
  gm.sq.est <- sum(var.est[K.step:T]) / (T - K.step + 1) ^ 2
  return(list(var.est = var.est, gm.sq.est = gm.sq.est))
}


var.fisher <- function(Yobs, Wobs, pT, p.lag, q.step) {
  # Description:
  #   This functions takes an observed outcome path and treatment path, and
  #   computes the variance bound given in (14)
  # Args:
  #   Yobs: A vector of length T, the observed outcome path.
  #   Wobs: A vector of treatment assignments
  #   pT: A numeric, the probability of assigning a unit to treatment.
  #   p.lag: A non-negative integer indicating the p-lag.
  #   q.step: A positive integer; indicating the number of steps.
  # Returns:
  #   var.est: The estimated variance bound at each time point.
  #   gm.sq.est: The estimate of gm (the average of the variances). 
  T <- length(Wobs)
  if (class(Wobs) == "factor"){
    # relable the factor levels
    levels(Wobs) <- c(0, 1)
    Wobs <- as.numeric(Wobs) - 1
  }
  # Turn pT into a vector
  if (length(pT) == 1) {
    pT.temp <- rep(pT, T)
    pT.temp[Wobs == 0] <- 1 - pT
    pT <- pT.temp
  }
  K.step <- p.lag + q.step
  if (K.step > 0) {
    binmat <- sapply(1:(2^(K.step + 1)), 
                   function(xx) {as.integer(intToBits(xx))[1:(K.step + 1)]})
    pinv <- sapply(1:ncol(binmat), function(xx) {1 / pTvec(pT, binmat[, xx])})
    pT.prod <- rowSums(pinv)
  } else {
    pT.prod <- pT
  }
  if (K.step == 0){
    var.est <- as.numeric(Yobs^2 / (pT * (1 - pT)))
  } else {
    var.est <- as.numeric(Yobs^2  * ((1 / 4)^(K.step)) * pT.prod)
  }
  gm.sq.est <- sum(var.est[K.step:T]) / (T - K.step) ^ 2
  return(list(var.est = var.est, gm.sq.est = gm.sq.est))
}


pTvec <- function(pT, treat.vec) {
  # Description:
  #   This function is used in the computation of the Fisher variance. 
  # Args:
  #   pT: The probability of receiving treatment at time T
  #   treat.vec: The treatment assignment vector
  # Returns:
  #   pT.out: A vector of length = length(pT) that gives the probability of
  #           observing the treat assignment specified by treat.vec at time t.
  T <- length(pT)
  k.step <- length(treat.vec)
  pT.mat <- matrix(1, nrow = k.step, ncol = T)
  pT.mat[1, ] <- c(treat.vec, rep(tail(treat.vec, 1), T - k.step)) * pT +
              (1 - c(treat.vec, rep(tail(treat.vec, 1), T - k.step))) * (1 - pT)
  if (k.step > 2) {
    for (jj in 2:(k.step - 1)) {
      Z <- c(treat.vec[1:(k.step - jj + 1)], 
            rep(tail(treat.vec[1:(k.step - jj + 1)], 1), T - k.step))
      pT.mat[jj, jj:T] <- 
          Z * pT[1:(T - jj + 1)] + (1 - Z) * (1 - pT[1:(T - jj + 1)])
    }
  }
  Z <- rep(treat.vec[1], T - k.step + 1)
  pT.mat[k.step, k.step:T] <- 
          (1 - Z) * (1 - pT[1:(T - k.step + 1)]) + Z * pT[1:(T - k.step + 1)]
  pT.out <- apply(pT.mat, 2, prod)
  return(pT.out)
}

combine.step.est <- function(Yobs, Wobs, pT, p.lag = 0, q.step = 0,
                             fisher.null = FALSE, normalize = FALSE) {
  # Description:
  #   This function computes the K-step estimator or the lagged-K step estimator
  #   as given in our paper for each unit and then return the combines estimate,
  #   given in equation (18).
  # Args:
  #   Yobs: A list of the the observed data of length N.
  #   Wobs: A list of the corresponding treatment assignments of length N.
  #   pT: Either a number giving the probability of receiving treatment at each
  #       time point, or a vector containing the probability of receiving
  #       treatment at each time poit.
  #   p.lag: A non-negative integer indicating the p-lag.
  #   q.step: A positive integer; indicating the number of steps.
  #   fisher.null: A logical, TRUE uses the fisher variance, FALSE uses the 
  #                variance bound.
  #   normalize: Logical, set true to use the normalized estimator.
  # Returns:
  #   est: The estimated value
  stopifnot(length(Yobs) == length(Wobs))
  if (class(pT) != "list") {
    pT <- lapply(1:length(Yobs), function(kk) pT)
  }

  N <- length(Yobs)
  est <- rep(NA, N)
  gm <- rep(NA, N)
  Ti <- sapply(1:N, function(xx) length(Yobs[[xx]]))
  for (kk in 1:N) {
    est[kk] <- est.ht(Yobs[[kk]], Wobs[[kk]], pT[[kk]], p.lag, q.step,
                      normalize)$est.mean
    if (!normalize){
      if (fisher.null) {
        gm[kk] <- var.fisher(Yobs[[kk]], Wobs[[kk]], pT[[kk]], p.lag,
                             q.step)$gm.sq.est
      } else {
        gm[kk] <- var.bound(Yobs[[kk]], Wobs[[kk]], pT[[kk]], p.lag,
                             q.step)$gm.sq.est
      } 
    } else {
      gm <- 1 / Ti
    }
  }
  gm.pool <- sum((1) / gm)
  nu <- sum(est / (gm / (1))) / gm.pool
  return(list(est = est, gm = gm, Ti = Ti, gm.pool = gm.pool, 
              nu = nu))
}


p.value.fisher.pool <- function(Yobs, Wobs, pT, p.lag = 0, q.step = 0, 
                                N.it = 1000, normalize = FALSE) {
  # Description:
  #   This functions takes an observed outcome path and treatment path, and
  #   computes a randomization based p-value for the ATE = delta.
  # Args:
  #   Yobs: A list of the the observed data of length N.
  #   Wobs: A list of the corresponding treatment assignments of length N.
  #   pT: Either a number giving the probability of receiving treatment at each
  #       time point, or a vector containing the probability of receiving
  #       treatment at each time poit.
  #   p.lag: A non-negative integer indicating the p-lag.
  #   q.step: A positive integer; indicating the number of steps.
  #   fisher.null: A logical, TRUE uses the fisher variance, FALSE uses the 
  #                variance bound.
  #   N.it: A numeric, the number of randomizations used to generate the p-value
  #   normalize: Logical, set true to use the normalized estimator.
  # Returns:
  #   p.val: The estimated p-value.
  #   rand.dist: The randomization distribution. 
  if (class(pT) != "list") {
    pT <- lapply(1:length(Yobs), function(kk) pT)
  }
  rand.dist <- rep(NA, N.it)
  p.val <- 0
  N <- length(Yobs)
  T <- sapply(1:N, function(xx) length(Yobs[[xx]]))
  est <- combine.step.est(Yobs, Wobs, pT, p.lag, q.step, TRUE, 
                          normalize = normalize)
  for(kk in 1:N.it) {
    cat("\r")
    cat(paste0("P-val is at iteration: ", kk))
    Wnew <- lapply(1:N, function(xx) rbinom(T[xx], size = 1, prob = pT[[xx]]))
    est.t <- combine.step.est(Yobs, Wnew, pT, p.lag, q.step, TRUE, 
                              normalize = normalize)
    rand.dist[kk] <- est.t$nu
    p.val <- p.val + ifelse(abs(rand.dist[kk]) >= abs(est$nu), 1, 0)
  }
  p.val <- p.val / N.it
  return(list(p.val = p.val, rand.dist = rand.dist, 
              nu.est = est$nu, gm.pool = est$gm.pool))
}



p.value.normal.pool <- function(Yobs, Wobs, pT, p.lag = 0, q.step = 0, 
                                ap = 0.05){
  # Description:
  #   This functions takes an observed outcome path and treatment path, and
  #   computes conservative p-value.
  # Args:
  #   Yobs: A vector of length T, the observed outcome path.
  #   W: A vector of treatment assignments
  #   pT: A numeric, the probability of assigning a unit to treatment.
  #   K.step: The degree of the estimator.
  #   lag: Logical, indicating the if this is the lagged k-step estimator (TRUE)
  #        or the k-step estimator (FALSE). Default is FALSE.
  # Returns:
  #   Z: The Z statistic.
  #   p.val: The estimated p-value.
  N <- length(Yobs)
  Ti <- sapply(1:N, function(xx) length(Yobs[[xx]]))
  est <- combine.step.est(Yobs, Wobs, pT, p.lag, q.step, fisher.null = FALSE)
  Z <- est$nu * est$gm.pool
  p.val <- 2 * min(1 - pnorm(Z), pnorm(Z))
  CI <- est$nu + c(- qnorm(1 - ap / 2) / est$gm.pool , 
                   qnorm(1 - ap / 2) / est$gm.pool)
  return(list(Z = Z, p.val = p.val, nu.est = est$nu, gm.pool = est$gm.pool,
              CI = CI))
}
