library(data.table)
library(dplyr)
library(tidyr)

gen.obs <- function(T = 500, tau = 0.8, phi = 0.7, pT = 0.5, delta = 0,
                    sigma = 1, K = NULL, sigma.full = NULL,
                    n.dist = "norm") {
  # Description
  #   This function generates an AR(1) time series with phi as the coefficient,
  #   then the observed outcome is generated by adding a constant 
  #   additive effect of tau and a cumulation effect of delta. 
  # Args:
  #   T: A number indicating the total number of time setps
  #   tau: A number indicating the constant additive treatment effect.
  #   phi: A number indicating the autoregressive coefficient for the AR
  #        process.
  #   pT: A number indicating the probability of being assigned to treatment.
  #   delta: A number indicating the cumulation effect.
  #   sigma: A number indicating the variance of the Gaussian.
  #   K: A number indicating the number of lags the treatment will have an 
  #      effect. 
  #   sigma.null: A vector of length T giving the errors. This is useful
  #               for generating the full science table - as the error terms
  #               are constant accross simulations
  # Returns:
  #   Yobs: The observed outcome path.
  #   W: The observed treatment path.
  #   sigma.full: The vector containing the errors. 
  
  W <- rbinom(n = T, size = 1, prob = pT)
  if (!is.null(K)) {
    W.mat <- matrix(NA, nrow = K, ncol = T)
    for (kk in 1:K) {
      W.mat[kk, ] <- c(rep(0, kk), W[1 : (T - kk)]) * delta ^ kk
    }
    W.sumed <- colSums(W.mat)
  } else {
    W.sumed <- rep(0, T)
  }
  mu <- tau * W + W.sumed
  if (is.null(sigma.full)) {
    if (n.dist == "t"){
      sigma.full <- rt(n = T, df = sigma)
    } else {
      sigma.full <- rnorm(T, 0, sigma)
    }
  }
  x <- mu + sigma.full
  Yobs <- stats::filter(x, phi, method = "rec")
  return(list(Yobs = Yobs, W = W, sigma.full = sigma.full))
}


gen.multiple <- function(N, T, tau.T, phi, pT, delta, sigma, K = NULL,
                         n.dist = "norm", p.lag, q.step,  normalize = FALSE, 
                         p.val.calc = TRUE, rand.pval = FALSE, same.sf = TRUE,
                         seed = 1) {
  # Description
  #   This functions computes multiple estimates of the k step estimator
  #   for the same errors using gen.obs. - Useful for measuring properties of
  #   the estimator over the randomization distribution.
  # Args:
  #   N: A number indicating the number of paths of the science table to try.
  #   T: A number indicating the total number of time setps
  #   tau.T: A number indicating the constant additive treatment effect.
  #   phi: A number indicating the autoregressive coefficient for the AR
  #        process.
  #   pT: A number indicating the probability of being assigned to treatment.
  #   delta: A number indicating the cumulation effect.
  #   sigma: A number indicating the variance of the Gaussian.
  #   K: A number indicating the number of lags the treatment will have an 
  #      effect. Default is NULL.
  #   n.dist: The distribution of the noise in the outcome model. Options are
  #           normal or t. Sigma is either the variance or df. 
  #   p.lag: A non-negative integer indicating the p-lag.
  #   q.step: A positive integer; indicating the number of steps.
  #   normalize: If rand.pval is true setting normalize to TRUE will use the 
  #              standardized statistic.
  #   p.val.calc: Logical indicating if p-values should be computed. Useful, 
  #               to set to false when just focusing on the distribution of 
  #               tau.
  #   rand.pval: A logical indicated if we should use the randomization p-values
  #             (true) or the normal approximation (FALSE)
  #   same.sf: If true uses the same noise distribution if false re-samples it
  #            every time.
  #   seed: The seed
  # Returns:
  #   tau: A vector of length N the corresponding tau,
  #   var.tau: A vector of length N of the variances of the tau vector
  #   var.Yobs: A vector of length N of the variance of the Yobs
  set.seed(seed)
  tau <- c()
  if (p.val.calc) {
    p.val <- c()
  } else {
    p.val <- NULL
  }
  obs <- gen.obs(T, tau, phi, pT, delta, sigma, K, n.dist = n.dist)
  if (same.sf){
    sf <- obs$sigma.full
  }
  if (p.val.calc) {
    if (rand.pval) {
      pn <- p.value.rand(Yobs = obs$Yobs, W = obs$W, pT = pT, p.lag = p.lag,
                         q.step = q.step, normalize = normalize)
      p.val[1] <- pn$p.val
    } else { 
      pn <- p.value.norm(Yobs = obs$Yobs, W = obs$W, pT = pT, p.lag = p.lag,
                         q.step = q.step)
      p.val[1] <- pn$p.val
    }
  }
  tau[1] <- pn$est
  for (kk in 2:N) {
    set.seed(kk + seed)
    if(kk %% 100 == 0) {
      cat(paste0("iteration: ", kk, "\n"))
    }
    if (same.sf){
      obs <- gen.obs(T, tau, phi, pT, delta, sigma, K, sf)
    } else {
      obs <- gen.obs(T, tau, phi, pT, delta, sigma, K)
    }
    if (p.val.calc) {
      if (rand.pval) {
        pn <- p.value.rand(Yobs = obs$Yobs, W = obs$W, pT = pT, p.lag = p.lag,
                         q.step = q.step, normalize = normalize)
        p.val[kk] <- pn$p.val
      } else {
        pn <- p.value.norm(Yobs = obs$Yobs, W = obs$W, pT = pT, p.lag = p.lag,
                         q.step = q.step)
        p.val[kk] <- pn$p.val
      }
    }
    tau[kk] <- pn$est
  }
  return(data.table(tau = tau, p.val = p.val))
}

gen.multiple.Z <- function(N, T, tau.T, phi, pT, delta, sigma, K, seed, nn, 
                           p.lag, q.step) {
  # Description
  #   This functions computes multiple estimates of the v estimator
  #   for the same errors using gen.obs. - Useful for measuring properties of
  #   the estimator over the randomization distribution.
  # Args:
  #   N: A number indicating the number of paths of the science table to try.
  #   T: A number indicating the total number of time setps
  #   tau.T: A number indicating the constant additive treatment effect.
  #   phi: A number indicating the autoregressive coefficient for the AR
  #        process.
  #   pT: A number indicating the probability of being assigned to treatment.
  #   delta: A number indicating the cumulation effect.
  #   sigma: A number indicating the variance of the Gaussian.
  #   K: A number indicating the number of lags the treatment will have an 
  #      effect. 
  #   sigma.null: A vector of length T giving the errors. This is useful
  #               for generating the full science table - as the error terms
  #               are constant accross simulations
  #   nn: New noise, should we use a new sample for the noise.
  # Returns:
  #   tau: A vector of length N the corresponding tau,
  #   var.tau: A vector of length N of the variances of the tau vector
  #   var.Yobs: A vector of length N of the variance of the Yobs
  set.seed(seed)
  Z <- c()
  var.Yobs <- c()
  p.val <- c()
  est <- c()
  obs <- gen.obs(T = T, tau = tau.T, phi = phi, pT = pT,
                     delta = delta, sigma = sigma, K = K)
  sf <- obs$sigma.full
  Z.est <- p.value.norm(obs$Yobs, obs$W, pT, p.lag, q.step)  
  est[1] <- Z.est$est
  p.val[1] <- Z.est$p.val
  Z[1] <- Z.est$Z
  var.Yobs <- sum(obs$Yobs ^ 2 / ((1-pT) * pT))/ T^2
  for (kk in 2:N) {
    set.seed(kk + seed)
    if(kk %% 100 == 0) {
      cat(paste0("iteration: ", kk, "\n"))
    }
    if (nn) {
      obs <- gen.obs(T = T, tau = tau.T, phi = phi, pT = pT,
                     delta = delta, sigma = sigma, K = K)
    } else {
      obs <- gen.obs(T = T, tau = tau.T, phi = phi, pT = pT,
                     delta = delta, sigma = sigma, K = K, sigma.full = sf)
    }
    Z.est <- p.value.norm(obs$Yobs, obs$W, pT, p.lag, q.step)
    p.val[kk] <- Z.est$p.val
    Z[kk] <- Z.est$Z
    est[kk] <- Z.est$est
  }
  return(list(est = est, Z = Z, p.val = p.val, var.Yobs = var.Yobs))
}




tau.T <- c(0, 0.25, 0.5, 0.75)
phi <- c(0.25, 0.5, 0.75)
pT <- c(0.1, 0.3, 0.5)
T <- c(100, 500, 1000)
resR <- data.table(tau = numeric(), p.val = numeric(), phi = numeric(),
                  tau.T = numeric(), T = numeric(), pT = numeric())


for (tT in tau.T) {
  for (ph in phi) {
    for (pTT in pT) {
      for (TT in T) {
        cat("Working on ") 
        cat(c(tT, ph, pTT, TT))
        cat("\n")
        resR.temp <- gen.multiple(N = 1000, T = TT, tau.T = tT, 
                                  phi = ph, pT = pTT,
                                  delta = 0.4, sigma = 1, K = NULL, 
                                  n.dist = "norm",
                                  p.lag = 0, q.step = 0,
                                  p.val.calc = TRUE, 
                                  rand.pval = TRUE, 
                                  normalize = FALSE, 
                                  same.sf = FALSE)
        resR.temp$phi <- ph
        resR.temp$tau.T <- tT
        resR.temp$T <- TT
        resR.temp$pT <- pTT
        resR <- rbind(resR, resR.temp)
      }
    }
  }
}

res.summaryR <- resR[, .(power = (sum(p.val <= 0.05) / 1000)), 
                   by = c("phi", "tau.T", "T", "pT")]
res.summaryR$phi <- factor(res.summaryR$phi)
res.summaryR$tau.T <- factor(res.summaryR$tau.T)
res.summaryR$T <- factor(res.summaryR$T)
res.summaryR$pT <- factor(res.summaryR$pT)


anova(lm(power ~ method * phi * tau.T * T * pT, data = res.full[tau.T != 0, ]))
anova(lm(power ~ method * phi * T * pT, data = res.full[tau.T == 0, ]))
res.summary[, mean(power), by = pT]
res.summary[, mean(power), by = phi]
res.summary[, mean(power), by = T]
res.summary[, mean(power), by = tau.T]

res.summary[, mean(power), by = c("tau.T", "pT")]
res.summary$method <- "norm"

res.summaryR$method <- "rand"
res.summaryR[, mean(power), by = pT]
res.summaryR[, mean(power), by = phi]
res.summaryR[, mean(power), by = T]
res.summaryR[, mean(power), by = tau.T]

res.summaryR[, mean(power), by = c("tau.T", "pT")]

res.full <- rbind(res.summary, res.summaryR)


dcast(res.full[tau.T != 0, round(mean(power), 3), by = c("pT", "T", "tau.T", "phi")], 
      phi ~ tau.T + T)
dcast(res.full[tau.T != 0, round(mean(power), 3), by = c("T", "tau.T", "phi")], 
      phi ~ tau.T + T)
dcast(res.full[tau.T == 0, round(mean(power), 2), by = c("pT", "method", "T")], 
      pT ~ method  + T)
dcast(res.full[tau.T == 0, round(mean(power), 2), by = c("pT", "method")], 
      pT ~ method )
dcast(res.full[tau.T != 0, round(mean(power), 3), by = c("pT", "method", "tau.T")], 
      pT + tau.T ~ method)
dcast(res.full[, mean(power), by = c("phi", "method")], 
      phi ~ method, value.var = "V1")
dcast(res.full[, mean(power), by = c("T", "method")], T ~ method)
dcast(res.full[, mean(power), by = c("tau.T", "method")], tau.T ~ method)

aa <- dcast(res.full[tau.T != 0, round(mean(power), 2), 
      by = c("tau.T", "pT", "method", "T")], 
      pT + tau.T  ~  method + T) 

sim.res <- gen.multiple.Z(N = 50000, T = 100, tau.T = 0.2, phi = 0.5, pT = 0.1,
                          delta = 0, sigma = 1, K = NULL, seed = 1, nn = FALSE, 
                          p.lag = 2, q.step = 0)
# sim.res$var.Yobs
# qqnorm(sim.res$Z)

dt.tau <- data.frame(tau = sim.res$est)
pp <- ggplot(data = dt.tau, aes(x = tau, y = ..density..)) + 
      geom_histogram(bins = 50, fill = "#577ECB", colour = "black", 
                     boundary = 0) + 
      geom_vline(xintercept = (0.2 * 0.5^2), colour = "#ED7D31") +
      theme_bw() + xlab(xlab(expression(tau))) + ylab("Density")
pp
ggsave(paste0("tau_2_100_2lag_0step_pT1_normal",".png"),
       width = 3.25, height = 2, units = "in")


sim.res <- gen.multiple.Z(N = 50000, T = 100, tau.T = 0.2, phi = 0.5, pT = 0.1,
                          delta = 0, sigma = 1, K = NULL, seed = 1, nn = FALSE, 
                          p.lag = 1, q.step = 0)
# sim.res$var.Yobs
# qqnorm(sim.res$Z)

dt.tau <- data.frame(tau = sim.res$est)
pp <- ggplot(data = dt.tau, aes(x = tau, y = ..density..)) + 
      geom_histogram(bins = 50, fill = "#577ECB", colour = "black", 
                     boundary = 0) + 
      geom_vline(xintercept = (0.2 * 0.5), colour = "#ED7D31") +
      theme_bw() + xlab(xlab(expression(tau))) + ylab("Density")
pp
ggsave(paste0("tau_2_100_1lag_0step_pT1_normal",".png"),
       width = 3.25, height = 2, units = "in")


sim.res <- gen.multiple.Z(N = 50000, T = 100, tau.T = 0.2, phi = 0.5, pT = 0.1,
                          delta = 0, sigma = 1, K = NULL, seed = 1, nn = FALSE, 
                          p.lag = 0, q.step = 1)
# sim.res$var.Yobs
# qqnorm(sim.res$Z)

dt.tau <- data.frame(tau = sim.res$est)
pp <- ggplot(data = dt.tau, aes(x = tau, y = ..density..)) + 
      geom_histogram(bins = 50, fill = "#577ECB", colour = "black", 
                     boundary = 0) + 
      geom_vline(xintercept = 0.2, colour = "#ED7D31") +
      theme_bw() + xlab(xlab(expression(tau))) + ylab("Density")
pp
ggsave(paste0("tau_2_100_0lag_1step_pT1_normal",".png"),
       width = 3.25, height = 2, units = "in")

sim.res <- gen.multiple.Z(N = 50000, T = 100, tau.T = 0.2, phi = 0.5, pT = 0.1,
                          delta = 0, sigma = 1, K = NULL, seed = 1, nn = FALSE, 
                          p.lag = 0, q.step = 0)
# sim.res$var.Yobs
# qqnorm(sim.res$Z)

dt.tau <- data.frame(tau = sim.res$est)
pp <- ggplot(data = dt.tau, aes(x = tau, y = ..density..)) + 
      geom_histogram(bins = 50, fill = "#577ECB", colour = "black", 
                     boundary = 0) + 
      geom_vline(xintercept = 0.2, colour = "#ED7D31") +
      theme_bw() + xlab(xlab(expression(tau))) + ylab("Density")
pp
ggsave(paste0("tau_2_100_0lag_0step_pT1_normal",".png"),
       width = 3.25, height = 2, units = "in")










